#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I implement the rdflib backend for TalesGraph.

This is used by `tales.TalesGraph.from_rdflib`
"""
from urlparse import urldefrag

from rdflib import BNode, ConjunctiveGraph, Literal, URIRef

from t4r.graph import TalesGraph
from t4r.mimetype import rdf_urlopen, guess_stream_mimetype
from t4r.path import RdfPath

class _RdflibTalesGraph(TalesGraph):
    def __init__(self, graph, context):
        TalesGraph.__init__(self, context)
        self._graph = graph

    def copy(self, context):
        """
        Make a copy of this tales_graph in another context.
        """
        return self.__class__(self._graph, context)

    def __str__(self):
        return "<graph %s>" % self._graph.identifier.n3()

    def size(self):
        return len(self._graph)

    def __iter__(self):
        """
        Iterate oveth the tuples of the graph.
        """
        make_tales_node = self._make_tales_node
        make_path = RdfPath._evaluated_path
        for t in self._graph:
            s = make_path(self, [make_tales_node(t[0])])
            p = make_path(self, [make_tales_node(t[1])])
            o = make_path(self, [make_tales_node(t[2])])
            yield dict(s=s, p=p, o=o, subject=s, predicate=p, object=o)

    def _eval_one_step_nodes(self, origin, property):
        encode = { "+": URIRef, "-": URIRef, "L": Literal, "_": BNode, }
        no = encode[origin[0]](*origin[1:])
        np = encode[property[0]](*property[1:])
        if property[0] == "-":
            r = self._graph.subjects(np, no)
        else:
            r = self._graph.objects(no, np)
        make_tales_node = self._make_tales_node
        for node in r:
            yield make_tales_node(node)

    def _make_tales_node(self, rdflib_node):
        u_node = unicode(rdflib_node)
        if isinstance(rdflib_node, URIRef):
            return ["+", u_node]
        elif isinstance(rdflib_node, Literal):
            datatype = rdflib_node.datatype
            if datatype is not None:
                datatype = unicode(datatype)
            return ["L", u_node, rdflib_node.language, datatype]
        else:
            return ["_", u_node]


type2format = {
    "application/rdf+xml": "xml",
    "text/n3": "n3",
    "application/turtle": "n3",
    "text/html": "rdfa",
}


class FynWrapper(object):
    """
    An object that is similar enough to a redland Model for _RedlandTalesGraph,
    but implementing "Follow your nose".
    """
    def __init__(self, t4r_options, follow, follow_inv):
        self.retrieved = set()
        self.graph = ConjunctiveGraph()
        self.t4r_options = t4r_options
        self.follow = [ URIRef(f) for f in follow ]
        self.follow_inv = [ URIRef(f) for f in follow_inv ]

    def __len__(self):
        return len(self.model)

    def __iter__(self):
        return iter(self.graph)

    def objects(self, ns, np):
        self._update_graph(ns)
        return self.graph.objects(ns, np)

    def subjects(self, np, no):
        self._update_graph(no)
        return self.graph.subjects(np, no)

    def _update_graph(self, node):
        allowed = self.t4r_options.get("allowed-requests")
        if allowed == 0: return
        if not isinstance(node, URIRef): return
        uri = urldefrag(str(node))[0]
        if uri in self.retrieved: return
        else:
            if allowed is not None:
                self.t4r_options["allowed-requests"] = allowed-1
            self.retrieved.add(uri)

        #print "===", "updating", node
        try:
            stream = rdf_urlopen(uri)
            content_type = guess_stream_mimetype(stream, uri)
            format = type2format.get(content_type)
            try:
                if format is None:
                    self.graph.parse(stream, uri)
                else:
                    self.graph.parse(stream, uri, format=format)
            finally:
                stream.close()
        except Exception:
            # too many exceptions can occur here to list them all;
            # (network or parsing related)
            # so we opt for a conservative behaviour.
            pass

        for f in self.follow:
            for n in list(self.graph.objects(node, f)):
                #print " ", "===", "following %s %s" % (f,n)
                self._update_graph(n)
        for f in self.follow_inv:
            for n in list(self.graph.subjects(f, node)):
                #print " ", "===", "following inverse %s %s" % (f,n)
                self._update_graph(n)
        # NB: it is better to wrap the calls to objects and subjects in lists,
        # because the API may not like to parse a file into the graph during an
        # iteration.

#
