#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I provide class QNameContext.
"""

from simpletal.simpleTALES import Context, ContextVariable, \
                                  PATHNOTFOUNDEXCEPTION

class QNameContext(Context):
    """
    A TALES context accepting qnames paths (relative and absolute).

    In order for a QNameContext to actually accept qnames paths, its method
    `set_graph_and_resource` must be called. This can not be done at `__init__`
    time because the graph needs the context to be constructed.
    """
    def __init__(self, options=None,
                 allowPythonPath=False, allowPrivateSymbols=False):
        Context.__init__(self, options, allowPythonPath)
        self.allowPrivateSymbols = allowPrivateSymbols
        g = self.globals = _QNameProxy(self)
        # the following is only to ensure that t4r always exists
        # but *may* be overridden, so nothing should be put in the dict here
        # (an example of overriding is subtemplate evaluation)
        g["t4r"] = {}

    def set_graph_and_resource(self, graph, resource):
        self.globals._set_resolver(resource) # handle relative paths
        self.addGlobal("", graph) # handle absolute paths
        self.addGlobal("graph", graph) # explicit graph variable
        self.addGlobal("resource", resource) # explicit graph variable

    def setLocal(self, name, value):
        if ":" in name:
            if name.startswith("t4rns"):
                raise TypeError("T4R namespaces should be declared as global")
            elif not name.startswith("t4r"):
                raise TypeError("QName variables are read-only: %s" % name)
        Context.setLocal(self, name, value)

    def traversePath(self, expr, canCall=1):
        """
        I override Context.traversePath to implement two features:
        * protecting private symbols (those starting with "_") if required
        * better indirection implementation (allow mutiple path items)
        """
        path = expr.split("/")
        if not self.allowPrivateSymbols:
            for p in path:
                if p and p[0] == "_":
                    raise PATHNOTFOUNDEXCEPTION
        for i,p in enumerate(path):
            if p and p[0] == "?":
                # manage dereference
                v = self.locals.get(p[1:]) or self.globals.get(p[1:])
                if v is not None:
                    if hasattr(v, "_TALES_dereference"):
                        v = v._TALES_dereference()
                    elif isinstance(v, ContextVariable):
                        v = v.value()
                    elif (callable (v)):
                        v = v()
                    path[i] = str(v)
        expr = "/".join(path)
        return Context.traversePath(self, expr, canCall)


class _QNameProxy():
    """
    A dict-like object accepting XQNames as read-only keys.

    Internally used by QnameContext for its global variables.
    """
    def __init__(self, context):
        self._context = context
        self._dict = context.globals

    def _resolve_item(self, key):
        raise ValueError("_set_resolver has not been invoked")

    def _qname2uriref(self, *skey):
        raise ValueError("_set_resolver has not been invoked")

    def _set_resolver(self, resolver):
        # this method is called by _resolve_item and _qname2uriref, to retrieve
        # the most appropriate resolver. Then, it replaces those methods by the
        # ones from the resover, for optimization.
        self._qname2uriref = resolver._qname2uriref
        self._resolve_item = resolver.__getitem__

    def __getitem__(self, key):
        skey = key.split(":")
        if 2 <= len(skey) <= 3 and not skey[0].startswith("t4r"):
            return self._resolve_item(key)
        else:
            return self._dict[key]

    def __setitem__(self, key, value):
        skey = key.split(":")
        if len(skey) >= 2:
            if not skey[0].startswith("t4r"):
                raise TypeError("QName variables are read-only: %s" % key)
            elif skey[1] in ["path", "exists", "nocall", "not", "string",
                             "python"]:
                raise TypeError("Do not use TALES keywords as namespaces: %s"
                                % skey[1])
        self._dict[key] = value

    def has_key(self, key):
        skey = key.split(":")
        len_skey = len(skey)
        if len_skey == 3:
            skey = skey[:2]
            len_skey = 2
        if len_skey == 2 and skey[0] != "t4rns":
            return self._qname2uriref(*skey) is not None
        else:
            return self._dict.has_key(key)

    def iteritems(self):
        return self._dict.iteritems()

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

