#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I define the class `TalesGraph` that implement the evaluation of `RdfPath`'s to
node lists.
"""
from decimal import Decimal
from itertools import chain
import logging
from urllib import urlencode, pathname2url
from urlparse import urljoin, urlunparse

from simpletal.simpleTALES import PathNotFoundException, PATHNOTFOUNDEXCEPTION

from t4r.path import RdfPath, RdfPathException, node_to_n3
from t4r.mimetype import guess_stream_mimetype, rdf_urlopen
import t4r.tal # do not import symbos, because of recursive import

class TalesGraph(object):
    """
    A TALES object representing an RDF graph.

    This is an abstract class that should be overridden for specific RDF
    backends. It uses `RdfPath` for parsing complex TALES paths.
    """

    # public common implementation

    @staticmethod
    def from_url(url, context):
        """
        Creates a TalesGraph from a standard URL.
        """
        try:
          from RDF import Model, Parser, RedlandError
          redland = True
        except ImportError:
          redland = None

        if redland:
            # use redland
            m = Model()
            p = Parser("guess")
            try:
                p.parse_into_model(m, url)
            except RedlandError:
                # if "guess" parser fail, try GRDDL
                p = Parser("grddl")
                p.parse_into_model(m, url)
            return TalesGraph.from_redland(m, context)
        else:
            # use rdflib
            f = rdf_urlopen(url)
            return TalesGraph.from_stream(f, context, url)

    @staticmethod
    def from_stream(stream, context, url=None):
        """
        Create a TalesGraph from a file-like object that must be either
        * a file-like object with an ``info`` method providing metadata (as
          returned by `urllib2.urlopen` or `t4r.make_typed_stream`) or
        * a real file object (as returned by ``open``).
        """
        if url is None:
            url = getattr(stream, "url", None)
        if url is None and hasattr(stream, "name"):
            url = urlunparse(
                ("file", None, pathname2url(stream.name), None, None, None))
        if url is None:
            url = "data://"

        try:
            from RDF import Model, Parser, RedlandError
            redland = True
        except ImportError:
            redland = None
            import rdflib

        if redland:
            # use redland
            m = Model()
            p = Parser("guess")
            try:
                p.parse_string_into_model(m, stream.read(), url)
            except RedlandError:
                p = Parser("grddl")
                p.parse_string_into_model(m, stream.read(), url)
            return TalesGraph.from_redland(m, context)

        else:
            # use rdflib
            from be_rdflib import type2format
            g = rdflib.ConjunctiveGraph(identifier=url)
            content_type = guess_stream_mimetype(stream, url)
            format = type2format.get(content_type)
            if format is not None:
                g.parse(stream, url, format=format)
            else:
                g.parse(stream, url) # let rdflib be clever
            return TalesGraph.from_rdflib(g, context)

    @staticmethod
    def from_rdflib(graph, context):
        """
        Creates a TalesGraph from an rdflib graph, linked to the given context.
        """
        from t4r.be_rdflib import _RdflibTalesGraph
        return _RdflibTalesGraph(graph, context)

    @staticmethod
    def from_redland(model, context):
        """
        Creates a TalesGraph from a redland model, linked to the given context.
        """
        from t4r.be_redland import _RedlandTalesGraph
        return _RedlandTalesGraph(model, context)

    @staticmethod
    def from_sparql_endpoint(url, context):
        """
        Creates a TalesGraph from a SPARQL endpoint URL.
        """
        raise NotImplementedError("TODO")

    @staticmethod
    def from_follow_your_nose(context):
        """
        Creates a follow your nose TalesGraph.
        """

        follow = [
            "http://www.w3.org/2000/01/rdf-schema#seeAlso",
            "http://www.w3.org/2002/07/owl#sameAs",
        ]
        follow_inv = [
            "http://www.w3.org/2002/07/owl#sameAs",
        ]

        try:
          import RDF
          redland = RDF
        except ImportError:
          redland = None

        if redland:
            from be_redland import FynWrapper
            w = FynWrapper(context.globals["t4r"], follow, follow_inv)
            r = TalesGraph.from_redland(w, context)
        else:
            from be_rdflib import FynWrapper
            w = FynWrapper(context.globals["t4r"], follow, follow_inv)
            r = TalesGraph.from_rdflib(w, context)
        r._fyn = True
        return r


    def __getitem__(self, key):
        """
        Return the TalesRdfNode corresponding to the qname.

        If the qname is not resolvable (see `_qname2uriref`), a KeyError is
        raised. Note however that if the qname resolves to a uri-ref, the
        corresponding node is returned *even* if the graph contains no triple
        about that resource.
        """
        return RdfPath.for_graph(self, key)

    def is_fyn(self):
        """
        Return whether this graph uses the "follow your nose" heuristic.
        """
        return self._fyn

    # protected common implementation

    def _qname2uriref(self, prefix, suffix):
        """
        Converts a qname (prefix:suffix) into a URI ref.

        The prefix of the qname is searched in the context variables of the
        form 't4rns:xxx'. If no such prefix is found, None is returned.
        """
        if prefix == "":
            turl = self._context.globals["t4r"].get("_template_url")
            if turl is None:
                log.warning("empty prefix in URL-less template")
                return None
            uri_prefix = urljoin(turl, ".")
        else:
            try:
                uri_prefix = self._context_evaluate("t4rns:%s" % prefix)
            except PathNotFoundException:
                log.warning("unknown prefix: %s", prefix)
                return None
        return uri_prefix + suffix

    def __init__(self, context):
        """
        Do not use it! Use `from_rdflib` or `from_sparql` instead.

        Only subclasses' __init__ should call this method.
        """
        self._context = context
        self._context_evaluate = context.evaluate
        self._fyn = False
        self._load = 0

    # public abstract methods
    # (some methods actually have a default implementation)

    def size(self):
        """
        Return the number of triples in this graph.
        """
        raise NotImplementedError("Abstract method")

    def __iter__(self):
        """
        Iter over the triples of the graph.

        This is equivalent to executing the following SPARQL query on the
        graph:

          SELECT ?subject, ?predicate, ?object
          WHERE { ?subject ?predicate ?object }
        """
        raise NotImplementedError("Abstract method")

    def copy(self, context):
        """
        Copy this TalesGraph in another context.
        """
        raise NotImplementedError("Abstract method")

    def _eval_path(self, path):
        """
        This method is the entry point of path evaluation (the return type
        depends on the kind of path, see below).

        It provides a default implementation, relying on all other `_eval_X`
        methods, the only abstract one being `_eval_one_step_nodes`. Hence,
        subclasses are only required to override the latter.

        For optimization issues, however, subclasses may decide to override
        any part of this method hierarchy. For this purpose, here is a
        documentation of the internal structure of an `RdfPath`.

        Basic path
        ==========

        A basic path is a path which evaluates into a list of RDF nodes. It
        means that it contains no data operator or binary operator. It can be
        recognized to the fact that its `_operator` attribute is `None`.

        The only interesting attribute in a basic path is `_steps`. It contains
        a list of lists of lists. Third level lists represent nodes or special
        path items (see below). Second level lists represent disjunction
        (sequences of ``or``-separated path items). See below for an example.

        Nodes
        -----

        RDF nodes are represented as follow:

        URI-ref: a 2-elements list containg the string "+" and the URI-ref
            itself as a unicode.
        Literal: a 4-elements list, whose first element is the string "L", the
            literal value as a unicode, the language tag if any as a unicode
            (or None), and the datatype URI if any as a unicode (or None).
        Blank nodes: a 2-elements list containing the string "_" and the blank
            node ID as a unicode.

        Special path items
        ------------------

        An `RdfPath` may contain other kinds of elements:

        Keyword: a 2-elements list containing the string "K" and the keyword as
            a unicode. Only node operators (see `RdfPath.NODE_OPERATORS`) can
            appear as keywords. The set `RdfPath.NODE_OPERATORS` contains all
            possible node operators.
        Inverse properties: a 2-elements list containing the string "-" and the
            URI-ref of the property as a unicode.

        NB: The first disjunction of the `_steps` only contain nodes, no
        special path items.

        Example
        -------

        The following TALES path::

            example:book1/foaf:made:-/or/dc:creator/datatype/or/rdf:type

        will be represented by the following structure::

            [
              [
                ["+", u"http://example.com/#book1"],
              ],
              [
                ["-", u"http://xmlns.com/foaf/0.1/made"],
                ["+", u"http://purl.org/dc/terms/creator"],
              ],
              [
                ["K", u"datatype"],
                ["+", u"http://www.w3.org/1999/02/22-rdf-syntax-ns#type"],
              ],
            ]

        Result of the evaluation
        ------------------------

        The result of the evaluation of a basic path must be an iterable of
        nodes, represented as defined above.

        Complex path
        ============

        A complex path is either a basic path followed by a data operator,
        or two basic paths separated by a binary operator. In both cases, the
        operator is contained by the `_operator` attribute, as a unicode.

        If the operator is unary (data operator), the corresponding basic path
        is stored in the `_steps` attribute.

        If the operator is binary, the left operand is stored in the `_steps1`
        attributes, while the right operand is stored in `_steps`.

        In any case, the `RdfPath` has a `_type` attribute, indicating the
        expected output type of the operator. It can be either one of the
        following python types: bool, int, unicode, or `RdfPath.UNKNOWN`
        (meaning that the actual type depends on the RDF data to be evaluated).
        The dicts `RdfPath.DATA_OPERATORS` and `RdfPath.BINARY_OPERATORS`
        contain all possible operators as their keys and the type of the
        corresponding operator as their values.

        Result of the evaluation
        ------------------------

        The result of the evaluation must be an instance of the corresponding
        `_type` (unless its is UNKNOWN, of course).

        Errors
        ======

        At any point, an error may be reported by returning None, and logging
        into "t4r.eval".

        """
        r = self._eval_steps(path._steps)
        operator = path._operator
        if operator is None:
            return r
        else:
            method = getattr(self, "_eval_op_%s" % operator)
            steps1 = path._steps1
            if steps1 is None:
                return method(r)
            else:
                r1 = self._eval_steps(steps1)
                r2 = r
                return method(r1, r2)

    def _eval_op_convert(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only convert a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        if node[0] != "L":
            log.error("can only convert literal")
            raise RDF_PATH_EXCEPTION
        converter = xsd_converters.get(node[3])
        if converter is None:
            log.error("can not convert datatype %s", node[3])
            raise RDF_PATH_EXCEPTION
        return converter(node[1])

    def _eval_op_id(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only get id from a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        typ = node[0]
        if typ == "_":
            return node[1]
        elif typ == "+":
            uri = node[1]
            hash = uri.rfind("#")
            slash = uri.rfind("/")
            colon = uri.rfind(":")
            cut = max(hash, slash, colon)
            return uri[cut+1:]
        else:
            log.error("can only get id from URIRef of Blank")
            raise RDF_PATH_EXCEPTION

    def _eval_op_isURIRef(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only test a single node")
            raise RDF_PATH_EXCEPTION
        return nodes[0][0] == "+"

    def _eval_op_isLiteral(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only test a single node")
            raise RDF_PATH_EXCEPTION
        return nodes[0][0] == "L"

    def _eval_op_isBlank(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only test a single node")
            raise RDF_PATH_EXCEPTION
        return nodes[0][0] == "_"

    def _eval_op_n3(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only convert a single node to N3")
            raise RDF_PATH_EXCEPTION
        return node_to_n3(self._context, nodes[0])

    def _eval_op_uri(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only get the URI of a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        if node[0] != "+":
            log.error("can only get the URI of a URIRef")
            raise RDF_PATH_EXCEPTION
        return node[1]

    def _eval_op_text(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only get the text of a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        if node[0] != "L":
            log.error("can only get the text of a literal")
            raise RDF_PATH_EXCEPTION
        return node[1]

    def _eval_op_lang(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only get the lang of a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        if node[0] != "L":
            log.error("can only get the lang of a Literal")
            raise RDF_PATH_EXCEPTION
        return node[2]

    def _eval_op_asPath(self, nodes):
        lst = []
        context = self._context
        for n in nodes:
            if n[0] == "+":
                n = node_to_n3(context, n)
                if n[0] == "<":
                    log.warning("no qname for %s, ignoring", n)
                else:
                    lst.append(n)
        return u"/or/".join(lst)

    def _eval_op_asInversePath(self, nodes):
        lst = []
        context = self._context
        for n in nodes:
            if n[0] == "+":
                n = node_to_n3(context, n)
                if n[0] == "<":
                    log.warning("no qname for %s, ignoring", n)
                else:
                    lst.append("%s:-" % n)
        return u"/or/".join(lst)

    def _eval_op_count(self, nodes):
        c = 0
        for n in nodes:
            c += 1
        return c

    def _eval_op_link(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only get the URI of a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        if node[0] != "+":
            log.error("can only get the URI of a URIRef")
            raise RDF_PATH_EXCEPTION
        uriref = node[1]
        fields = self._context.globals.get("t4r", {}).get("fields")
        if fields is None:
            return None
        ts = fields.get("ts", [""])[0]
        tu = fields.get("tu", [""])[0]
        tt = fields.get("tt", [""])[0]
        gs = fields.get("gs", [""])[0]
        gu = fields.get("gu", [""])[0]
        gt = fields.get("gt", [""])[0]
        r = fields.get("r", [""])[0]
        if r:
            # this is a resource template
            r = uriref
        else:
            # this is a graph template
            gu = uriref
            gs = "" # mutually exclusive with gu
            gt = "" # safer to guess, since the graph has changed
            r = "" # makes more sense, since
        args = [("ts", ts), ("tu", tu), ("tt", tt)]
        if gs or gu:
            args.extend([("gs", gs), ("gu", gu), ("gt", gt)])
        if r:
            args.append(("r", r))
        return u"?" + urlencode(args)

    def _eval_op_contains(self, nodes1, nodes2):
        nodes2 = list(nodes2)
        for n1 in nodes1:
            if n1 in nodes2:
                return True
        return False

    def _eval_op_renderWith(self, nodes1, nodes2):
        nodes1 = list(nodes1)
        if len(nodes1) != 1:
            log.error("can only render a single node")
            raise RDF_PATH_EXCEPTION
        node1 = nodes1[0]
        if node1[0] != "+":
            log.error("can only render a URIRef")
            raise RDF_PATH_EXCEPTION
        nodes2 = list(nodes2)
        if len(nodes2) != 1:
            log.error("can only render with a single node")
            raise RDF_PATH_EXCEPTION
        node2 = nodes2[0]
        if node2[0] != "+":
            log.error("can only render with a URIRef")
            raise RDF_PATH_EXCEPTION

        # we pass the *same* dict as `t4r`, so that 'allowed-requests' and
        # '_ctemplates' are inherited and can be updated in subtemplate
        globals = { "t4r": self._context.globals["t4r"] }
        rendered = t4r.tal.render_template(node2[1], node1[1], self, globals)
        return rendered.read().decode("utf-8")

    def _eval_steps(self, steps):
        if len(steps) == 1:
            return steps[0]
        else:
            nodes = list(self._eval_steps(steps[:-1]))
            return chain (*(
                self._eval_one_step(nodes, step) for step in steps[-1]
            ))

    def _eval_one_step(self, nodes, step):
        if step[0] is "K":
            return getattr(self, "_eval_op_%s" % step[1])(nodes)
        else:
            return chain(*(
                self._eval_one_step_nodes(n, step) for n in nodes
            ))

    def _eval_op_datatype(self, nodes):
        for n in nodes:
            if n[0] == "L":
                datatype = n[3]
                if datatype is not None:
                    yield ["+", datatype]

    def _eval_op_any(self, nodes):
        try:
            yield iter(nodes).next()
        except StopIteration:
            raise PATHNOTFOUNDEXCEPTION

    def _eval_op_URIRefs(self, nodes):
        for n in nodes:
            if n[0] is "+":
                yield n

    def _eval_op_literals(self, nodes):
        for n in nodes:
            if n[0] is "L":
                yield n

    def _eval_op_blanks(self, nodes):
        for n in nodes:
            if n[0] is "_":
                yield n

    def _eval_op_iterSeq(self, nodes):
        for n in nodes:
            i = 1
            cont = True
            while cont:
                cont = False
                for p in self._eval_one_step_nodes(n, ["+", "%s_%s"%(RDF,i)]):
                    cont = True
                    yield p
                i += 1

    def _eval_op_iterList(self, nodes):
        return chain(*(
            self._eval_op_iterList_1(n) for n in nodes
        ))

    def _eval_op_iterList_1(self, node):
        if node != RDF_NIL:
            firsts = list(self._eval_one_step_nodes(node, RDF_FIRST))
            rests = list(self._eval_one_step_nodes(node, RDF_REST))
            lf = len(firsts); lr = len(rests)
            if lf > 1 or lr > 1 or lf != lr:
                log.error("Ill-formed list %s", node[1])
                raise RDF_PATH_EXCEPTION
            if lf == 1:
                return chain (firsts, self._eval_op_iterList_1(rests[0]))
        return []

    def _eval_op_ns(self, nodes):
        nodes = list(nodes)
        if len(nodes) != 1:
            log.error("can only get namespace of a single node")
            raise RDF_PATH_EXCEPTION
        node = nodes[0]
        if node[0] != "+":
            log.error("can only get namespace of a URIRef")
            raise RDF_PATH_EXCEPTION
        uri = node[1]
        hash = uri.rfind("#")
        slash = uri.rfind("/")
        colon = uri.rfind(":")
        cut = max(hash, slash, colon)
        yield ["+", uri[:cut+1]]

    def _eval_op_selectLang(self, nodes):
        t4r = self._context.globals.get("t4r", {})
        al = t4r.get(".parsed-accept-language")
        if al is None:
            hal = t4r.get("accept-language")
            if hal is None:
                return self._eval_op_literals(self._eval_op_literals(nodes))
            else:
                al = t4r[".parsed-accept-language"] = parse_accept_language(hal)

        found = {}
        for n in self._eval_op_literals(nodes):
            found.setdefault(("*",), []).append(n)
            lang = n[2]
            if lang:
                lang = tuple(lang.split("-"))
                found.setdefault(lang, []).append(n)
                if len(lang) == 2:
                    found.setdefault((lang[0], None), []).append(n)
            else:
                found.setdefault((), []).append(n)

        for i in al:
            r = found.get(i) or (len(i) == 1 and found.get(i+(None,)))
            if r:
                return r
        return []

    def _eval_one_step_nodes(self, origin, property):
        """
        Iter over the third components of matching RDF triples.

        `origin` is a node, `property` is either a node or an inverse property.
        """
        raise NotImplementedError("Abstract method")

#

def parse_accept_language(al):
    return [ tuple(i.split(";")[0].strip().split("-")) for i in al.split(",") ]

RDF_PATH_EXCEPTION = RdfPathException()

RDF= "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
RDF_FIRST = [ "+", "%sfirst" % RDF ]
RDF_REST = [ "+", "%srest" % RDF ]
RDF_NIL = [ "+", "%snil" % RDF ]

XSD = "http://www.w3.org/2001/XMLSchema#"

xsd_converters = {
    "%sint" % XSD: int,
    "%sinteger" % XSD: int,
    "%sfloat" % XSD: float,
    "%sdouble" % XSD: float,
    "%sdecimal" % XSD: lambda v: ("." in v) and Decimal(v) or int(v),
    "%sstring" % XSD: unicode,
    None: unicode,
}

log = logging.getLogger("t4r.path.eval")

#
