#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I provide method for parsing and serializing mimetypes, as well as guessing
mimetype from a stream.
"""
from mimetypes import guess_type
from os.path import abspath
from urllib import pathname2url
from urllib2 import urlopen, Request
from urlparse import urldefrag, urlunparse


RDF_MIMETYPES = (
    "application/rdf+xml",
    "application/turtle",
    "text/n3",
)

def parse_mimetype(mimetype):
    """
    I return a tuple of the form: (type, subtype, params) where params is a
    dict.
    Based on a code by joe@bitworking.org on redfoot.net .
    """
    parts = mimetype.split(";")
    part0 = parts[0]
    slash = part0.find("/")

    if slash >= 0:
        type_ = part0[:slash]
        subtype = part0[slash+1:]
    else:
        type_ = part0
        subtype = "*"

    params = [ p.split("=", 1) for p in parts[1:] ]
    params = dict( (k.strip(), v.strip()) for k,v in params )

    return (type_, subtype, params)

def unparse_mimetype(type_, subtype, params):
    parts = ["%s/%s" % (type_, subtype)]
    parts.extend( "%s=%s" % i for i in params.iteritems() )
    return ";".join(parts)

def strip_params(mimetype):
    """
    Return only the type/subtype part of the mimetype.
    """
    return mimetype.split(";", 1)[0]

def guess_stream_mimetype(stream, url=None):
    """
    Guess the mimetype of a stream (file-like) object.

    It is required that the stream has an 'info' method (as returned by urlopen)
    or is an instance of the 'open' builtin.
    """
    info = getattr(stream, "info", None)
    if info is not None:
        url = url or getattr(stream, "url", None)
        type_ = info().getheader("content-type")
    else:
        if url is None:
            url = pathname2url(abspath(stream.name))
            url = urlunparse(("file", None, url, None, None, None))
        type_ = guess_type(url)[0] or "application/octet-stream"

    if strip_params(type_) not in RDF_MIMETYPES and url is not None:
        url = urldefrag(url)[0].lower()
        suffix2format = [
            (".n3", "text/n3"),
            (".ttl", "application/turtle"),
            (".rdf", "application/rdf+xml"),
        ]
        for suffix, guessed in suffix2format:
            if url.endswith(suffix):
                type_ = guessed
                break

    return type_

def rdf_urlopen(url, accept=""):
    """
    A useful wrapper for urlopen, explicitly requesting RDF mimetypes.
    """
    global _accept
    return urlopen(Request(url, headers={"Accept": _accept}))

_accept = ",".join(RDF_MIMETYPES + ("text/html",))

