#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I define the class `RdfPath` that implement the syntactic part of rdf paths.
"""
from copy import deepcopy
import logging

from simpletal.simpleTALES import ContextVariable, PathNotFoundException


class RdfPath(ContextVariable):

    def __init__(self):
        """Do not use. Use `for_graph` or `_make_mutable_copy` instead."""
        ContextVariable.__init__(self, self)

    @classmethod
    def for_graph(cls, graph, first):
        path = cls()
        path._graph = graph
        path._context = graph._context
        path._mutable = True
        path._steps = [[]]
        path._steps1 = None
        path._operator = None
        path._type = NODES
        path._state = path._initial(path, first)
        return path

    @classmethod
    def for_resource(cls, graph, uriref):
        return cls._evaluated_path(graph, [["+", unicode(uriref)]])

    @classmethod
    def _evaluated_path(cls, graph, value):
        path = cls()
        path._graph = graph
        path._context = graph._context
        path._mutable = False
        path._steps = [list(value)]
        path._steps1 = None
        path._operator = None
        path._type = NODES
        path._state = path._root1
        return path

    def _make_mutable_copy(self):
        path = self.__class__()
        path._graph = self._graph
        path._context = self._context
        path._mutable = True
        path._state = self._state
        path._steps = deepcopy(self._steps)
        path._steps1 = deepcopy(self._steps1)
        path._operator = self._operator
        path._type = self._type
        return path

    def __getitem__(self, step):
        """
        Invoked by the TALES parser to add another step.

        It implements a state automaton working as follows.

        The `_state` attribute is a static method (hence `self` must be
        explicitly passed to this method) accepting as parameters:
         * the path instance
         * the string token to be consummed by the automaton
        It must return either `None` (failure) or the new state.
        Furthermore, every state method has a `terminal` attribute indicating
        whether this state is a valid terminal state.

        The class has a nufmber of state static methods (decorated with the
        @state() decorator).
        """
        if self._mutable:
            path = self
        else:
            path = self._make_mutable_copy()
        r = path._state(path, step)
        if r is None:
            log.error("unrecognized step %s", step)
            # _state == None will be handlered by value()
        path._state = r
        return path

    def value(self, currentPath=None):
        """
        Invoked by the TALES parser at each step.

        This method does nothing during intermediate steps, but computes the
        result at the very last step of the path.
        """
        if currentPath is None: return self
        index, pathList = currentPath

        state = self._state
        _type = self._type
        # first handle syntactical error
        if state is None:
            subpath = "/".join(pathList[:index])
            return RdfPathError("invalid path step: %s" % subpath)
        # must finish if either
        # - the path is complete
        # - a *unary* operator has been used,
        #   returning something else than a path
        if len(pathList) > index \
        and (_type is NODES or self._steps1 is not None):
            return self
        # this is the case, so let us evaluate the path to a node list
        if state is self._root1:
            self._mutable = False
            return self
        elif not state.terminal:
            log.error("terminated on non-terminal state: %s", self._state)
            return RdfPathError("unfinished path: %s" % ("/".join(pathList),))
        else:
            assert self._mutable # a non evaluated complex path is mutable
            r = self._graph._eval_path(self)
            if _type is not NODES:
                assert _type is UNKNOWN or isinstance(r, _type), \
                       "%r has not correct type %r" % (r, _type)
                return r
            else:
                r = list(r)
                self._mutable = False
                self._steps = [list(r)]
                self._steps1 = None
                self._operator = None
                self._state = self._root1
                return self

    def __iter__(self):
        assert self._state == self._root1
        g = self._graph
        cache = set()
        for i in self._steps[0]:
            index = tuple(i) # a list is not hashable, a tuple is
            if index not in cache:
                cache.add(index)
                yield RdfPath._evaluated_path(g, [i])

    def __str__(self):
        assert self._state == self._root1
        c = self._graph._context
        display = c.locals.get("t4r:display")
        if display is None:
            display = c.globals.get("t4r:display")
        nodes = self._steps[0]
        lnodes = len(nodes)
        if display == "strict" and lnodes != 1:
            return "{%s nodes}" % lnodes
        else:
            if lnodes == 0: return ""
            node = nodes[0]
            if node[0] == "_":
                return "_:%s" % node[1]
            else:
                return node[1]

    def __nonzero__(self):
        assert self._state == self._root1
        return len(self._steps[0])

    def state(terminal=False): # local method decorator
        """
        Decorate a method as a state, possibly a terminal one.

        Makes the method static (because it could be called on another
                instance).
        """
        def wrapper(f):
            f.terminal = terminal
            f = staticmethod(f)
            return f
        return wrapper

    @state()
    def _initial(self, step):
        return self._add_qname(step, self._root1, append=False)

    @state(terminal=True)
    def _root1(self, step):
        r = self._add_or(step, self._initial) \
            or self._add_node_op(step, self._prop1) \
            or self._add_xqname(step, self._prop1) \
            or self._add_data_op(step, self._data) \
            or self._add_binary_op(step, self._pending_binary_operator)
        return r

    @state(terminal=True)
    def _prop1(self, step):
        return self._add_or(step, self._pending_or1) \
            or self._add_node_op(step, self._prop1) \
            or self._add_data_op(step, self._data) \
            or self._add_xqname(step, self._prop1) \
            or self._add_binary_op(step, self._pending_binary_operator)

    @state()
    def _pending_or1(self, step):
        return self._add_xqname(step, self._prop1, append=False)

    @state(terminal=True)
    def _data(self, step):
        assert 0, "This should not happen"
        # indeed, self.value() should have been invoked and returned the data
        return None

    @state()
    def _pending_binary_operator(self, step):
        return self._add_qname(step, self._root2, append=False)

    @state(terminal=True)
    def _root2(self, step):
        return self._add_or(step, self._pending_binary_operator) \
            or self._add_node_op(step, self._prop2) \
            or self._add_xqname(step, self._prop2)

    @state(terminal=True)
    def _prop2(self, step):
        return self._add_or(step, self._pending_or2) \
            or self._add_node_op(step, self._prop2) \
            or self._add_xqname(step, self._prop2)

    @state()
    def _pending_or2(self, step):
        return self._add_xqname(step, self._prop2, append=False)

    del state # we do not need the decorator anymore

    def _add_qname(self, step, next_state, append=True):
        splitted = step.split(":")
        if len(splitted) == 2:
            uriref = self._qname2uriref(*splitted)
            if uriref is not None:
                #import pdb; pdb.set_trace()
                if append:
                    lst = []
                    self._steps.append(lst)
                else:
                    lst = self._steps[-1]
                lst.append(["+", uriref])
                return next_state
            else:
                log.error("qname %s has undeclared namespace prefix", step)
                raise RdfPathError("undeclared namespace: %s" % splitted[0])
        else:
            return None

    def _add_xqname(self, step, next_state, append=True):
        splitted = step.split(":")
        len_splitted = len(splitted)
        if 2 <= len_splitted <= 3:
            uriref = self._qname2uriref(*splitted[:2])
            if uriref is None:
                log.error("qname %s has undeclared namespace prefix", step)
                raise RdfPathError("undeclared namespace: %s" % splitted[0])
            if len_splitted == 3:
                suffix = splitted[2]
                if suffix == "-":
                    typed_step = ["-", uriref]
                else:
                    log.error("unknown xqname suffix: %s", suffix)
                    raise RdfPathError("unknown xqname suffix: %s" % suffix)
            else: # len_splitted == 2
                    typed_step = ["+", uriref]
            if append:
                lst = []
                self._steps.append(lst)
            else:
                lst = self._steps[-1]
            lst.append(typed_step)
            return next_state
        else:
            return None

    def _add_or(self, step, next_state):
        if step == "or":
            return next_state
        else:
            return None

    def _add_node_op(self, step, next_state):
        if step in NODE_OPERATORS:
            self._steps.append([["K", step]])
            return next_state
        else:
            return None

    def _add_data_op(self, step, next_state):
        if step in DATA_OPERATORS:
            self._operator = step
            self._type = DATA_OPERATORS[step]
            return next_state
        else:
            return None

    def _add_binary_op(self, step, next_state):
        if step in BINARY_OPERATORS:
            self._operator = step
            self._steps1 = self._steps
            self._steps = [[]]
            self._type = BINARY_OPERATORS[step]
            return next_state
        else:
            return None

    def _qname2uriref(self, *args):
        """
        Shortcut to `self._graph._qname2uriref`. Used by
        `context.QNameContext`.
        """
        m = self.__qname2uriref
        if m is None:
            m = self.__qname2uriref = self._graph._qname2uriref
        return m(*args)
    __qname2uriref = None


class RdfPathException(PathNotFoundException):
    """
    Raised whenever a path expression is *semantically* incorrect, i.e. when it
    can not be evaluated.

    :see-also: `RdfPathError`
    """
    pass


class RdfPathError(Exception, ContextVariable):
    """
    I am the result of a *syntactically* ill-formed path.

    This is better than throwing an `RdfPathException`, because the later will
    be silently ignored by TAL, and be rendered as an empty string. This is
    useful to build robust templates when the underlying data is not as
    expected.

    A syntax error, however, is different: this is a formal bug in the
    template. Hence, it is better to have this error clearly visible in the
    template so as to be easily diagnosed and fixed.
    """
    def __init__(self, msg):
        ContextVariable.__init__(self, self)
        self.msg = msg

    def __eq__(self, other):
        # useful for tests
        return isinstance(other, RdfPathError) and self.msg == other.msg

    def __repr__(self):
        return "RdfPathError(%r)" % self.msg

    def __str__(self):
        return "{RDF-Path ERROR: %s}" % self.msg

    def __getitem__(self, key):
        return self


def node_to_n3(context, node):
    ntype = node[0]
    if ntype == "+":
        uri = node[1]
        for prefix, nsuri in context.globals.iteritems():
            if prefix.startswith("t4rns:") and uri.startswith(nsuri):
                return "%s:%s" % (prefix[6:], uri[len(nsuri):])
        return "<%s>" % uri
    elif ntype == "L":
        val, language, datatype = node[1:]
        val = '"%s"' % val.replace('"', '\\"')
        if language is not None:
            return "%s@%s" % (val, language)
        elif datatype is not None:
            return "%s^^%s" % (val, node_to_n3(context, ["+", datatype]))
        else:
            return val
    else: # ntype == "_"
        return "_:%s" % node[1]

# pseudo-types constants
NODES = object()
UNKNOWN = object()
# NB: about UNKNOWN
# It means that the type of the value will be determined at evaluation time.
# However, this type can *not* be NODES, only a python datatype.

NODE_OPERATORS = set([
    "datatype",
    "any", # takes the first one, but most sequences are unordered
           # also raises PATHNOTFOUNDEXCEPTION if the path is empty
    "URIRefs",
    "literals",
    "blanks",
    "iterSeq", # keeps the order
    "iterList", # keeps the order
    "ns",
    "selectLang",
])

DATA_OPERATORS = {
    "convert": UNKNOWN, # see documentation for supported types
    "id": unicode,
    "isURIRef": bool,
    "isLiteral": bool,
    "isBlank": bool,
    "n3": unicode,
    "uri": unicode,
    "text": unicode,
    "lang": UNKNOWN, # unicode or None
    "asPath": unicode,
    "asInversePath": unicode,
    "count": int,
    "link": UNKNOWN, # unicode (url-encoded ascii) or None
}

BINARY_OPERATORS = {
    "contains": bool, # or intersects?
    "renderWith": unicode,
}

log = logging.getLogger("t4r.path.parse")
