#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I allow to manage correspondance between mimetypes and template species.

There are three species of templates:
* HTML
* XML
* OTHER

This module provides method `guess_species` which, given a mimetype, tries to
determine the correct species. It uses a resgister of common mimetypes, and
some heuristics on the structure of the mimetype itself.

If `guess_species` does not make the correct guess, it is possible to augment
and query the register, respectively with method `register_species` and
`get_species`.
"""

from t4r.mimetype import strip_params

HTML = 0
XML = 1
OTHER = 2

_register = {}

def register_species(mimetype, species):
    """
    See module docstring.
    """
    assert species in set([HTML, XML, OTHER])
    mimetype = strip_params(mimetype)
    current = _register.get(mimetype)
    if current is not None and current is not species:
        raise ValueError("%s already registered as %s"
                         % (mimetype, _S2S[current]))
    _register[mimetype] = species

def unregister_species(mimetype):
    mimetype = strip_params(mimetype)
    del _register[mimetype]

def get_species(mimetype):
    """
    See module docstring.
    """
    mimetype = strip_params(mimetype)
    return _register.get(mimetype)

def guess_species(mimetype):
    """
    See module docstring.
    """
    mimetype = strip_params(mimetype)
    guess =  _register.get(mimetype)
    if guess is None:
        if mimetype.endswith("+xml"):
            guess = XML
        else:
            guess = OTHER
    return guess


_S2S = { HTML: "HTML", XML: "XML", OTHER:"OTHER" }

register_species("text/html", HTML)
register_species("text/xml", XML)
register_species("application/xml", XML)
