#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I provide mechanism for template rendering.
"""

import logging
from mimetools import Message
import re
from StringIO import StringIO
import sys
import traceback
from urllib2 import urlopen
from urlparse import urldefrag

from simpletal.simpleTAL import compileXMLTemplate, compileHTMLTemplate, \
                                XMLTemplate

from t4r.species import guess_species, HTML, OTHER
from t4r.mimetype import parse_mimetype, unparse_mimetype, \
                         guess_stream_mimetype
from t4r.path import RdfPath
from t4r.graph import TalesGraph
from t4r.context import QNameContext

def render_template(template, resource, graph=None, globals=None,
                    logstream=None):
    """
    Render a template.

    Both `template` and `graph` (if provided) can be either
    * a file-like object with an ``info`` method providing metadata (as
      returned by `urllib2.urlopen` or `t4r.make_typed_stream`),
    * a real file object (as returned by ``open``), or
    * an instance of basetring representing a URL where the template or graph,
      respectively, can be downloaded.

    `resource`must be a URI. At least one of `resource` and `graph`
    must be provided.

    If `graph` is not provided, the resource URI is considered as the URL for
    retrieving the RDF data. Else, the resource URI is merely used as an
    identifier, and the RDF data is retrieved from the graph.

    Optionally, a dict `globals` can be passed to be added as global variables
    in the TAL context.

    Optionally, a file-like object `logstream` can be passed to receive log
    messages from the t4r logger.

    The result is a file-like object containing the rendered template. Note
    that such an object is also iterable, hence suitable as a WSGI result.
    """
    if graph is None:
        graph = urldefrag(resource)[0]
    context = make_context(graph, resource, globals)
    t4r_options = context.globals["t4r"]
    if logstream is not None:
        t4r_options["_logstream"] = logstream

    ctemplates = t4r_options.setdefault("_ctemplates", {})

    if isinstance(template, basestring):
        turl = t4r_options["_template_url"] = template
        tstream = urlopen(template)
    else:
        turl = None
        tstream = template

    if turl in ctemplates:
        ctemplate, parsed_ttype, encoding, species = ctemplates[turl]
    else:
        allowed = t4r_options.get("allowed-requests")
        if allowed is not None:
            if allowed == 0 and not turl.startswith("file:"):
                return StringIO()
            t4r_options["allowed-requests"] = allowed-1

        ttype = guess_stream_mimetype(tstream, turl)
        species = guess_species(ttype)
        parsed_ttype = parse_mimetype(ttype)
        encoding = parsed_ttype[2].get("charset")
        ctemplate = compile_template(tstream, encoding, species)
        ctemplates[turl] = (ctemplate, parsed_ttype, encoding, species)

    is_subtemplate = isinstance(graph, TalesGraph)
    out = process_tal(context, ctemplate, is_subtemplate)
    if species is OTHER:
        out = non_xml_filter(out)
    if encoding or species is HTML:
        parsed_ttype[2]["charset"] = "utf-8"
    add_info(out, unparse_mimetype(*parsed_ttype))
    return out

def add_info(stringio, content_type):
    info = Message(EMPTY_FILE)
    info["content-type"] = content_type
    info["content-length"] = str(stringio.len)
    stringio.info = lambda: info

def make_context(graph, resource, globals=None):
    """
    Parameter `graph` is the one passed to `render_graph_template` or
    `render_resource_template`.
    """
    c = QNameContext(resource)
    if isinstance(graph, TalesGraph):
        tg = graph.copy(c)
    elif isinstance(graph, basestring):
        global _sparql
        if graph == "fyn:":
            tg = TalesGraph.from_follow_your_nose(c)
        elif _sparql.search(graph):
            tg = TalesGraph.from_sparql_endpoint(graph, c)
        else:
            tg = TalesGraph.from_url(graph, c)
    else:
        tg = TalesGraph.from_stream(graph, c)

    r = RdfPath.for_resource(tg, resource)
    c.set_graph_and_resource(tg, r)

    if globals:
        for k,v in globals.iteritems():
            c.addGlobal(k,v)

    return c

_sparql = re.compile("^sparql:|/sparql/?$")

def compile_template(template_stream, encoding, species):
    if species is OTHER:
        wrapped_stream = StringIO()
        if encoding:
            wrapped_stream.write(
                '<?xml version="1.0" encodind="%s"?>' % encoding)
        wrapped_stream.write(
            "<non-xml xmlns:tal='http://xml.zope.org/namespaces/tal'>\n")
        wrapped_stream.write(template_stream.read())
        wrapped_stream.write("</non-xml>\n")
        wrapped_stream.seek(0)
        template_stream = wrapped_stream

    try:
        if species is HTML:
            if encoding is None: encoding = "ascii"
            return compileHTMLTemplate(template_stream, encoding)
        else:
            return compileXMLTemplate(template_stream)
    except Exception, e:
        raise T4rTemplateParsingError(e)

def process_tal(context, compiled_template, is_subtemplate):
    """
    Render the template with the given context into a StringIO, and return
    that StringIO.
    """
    logstream = context.globals["t4r"].get("_logstream")
    if logstream is None:
        logstream = StringIO()
    handler = logging.StreamHandler(logstream)
    handler.setFormatter(logging.Formatter('%(name)s %(levelname)s %(message)s'))
    log = logging.getLogger("t4r")
    log.addHandler(handler)
    log.setLevel(logging.WARNING)

    out = StringIO()
    try:
        kw = {}
        if isinstance(compiled_template, XMLTemplate):
            kw["suppressXMLDeclaration"] = is_subtemplate
        compiled_template.expand(context, out, outputEncoding="utf-8", **kw)
    except Exception, e:
        print >>sys.stderr, traceback.format_exc(e)
    finally:
        log.removeHandler(handler)

    out.seek(0)
    return out

def non_xml_filter(lines):
    """
    Drops the first two lines and last line of the iterable lines.

    (assuming lines has at least 3 lines)
    """
    out = StringIO()
    lines.next() # drop <?xml> declaration
    lines.next() # drop opening tag
    previous = lines.next()
    for l in lines:
        out.write(previous)
        previous = l
    # the last line is the closing tag
    out.seek(0)
    return out

class _EmbeddingException(Exception):
    def __str__(self):
        return "%s: %s" % (self.args[0].__class__.__name__, self.args[0])

class T4rTemplateParsingError(_EmbeddingException): pass

class T4rGraphParsingError(_EmbeddingException): pass

EMPTY_FILE = StringIO()

#

