#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

"""
I provide the WSGI application
"""

import cgi
from os.path import abspath, exists, join
from urllib import pathname2url
from urlparse import urljoin, urlparse, urlunparse

from t4r import render_template, make_typed_stream

def engine(env, start_response):
    """
    A WSGI application doing the actual rendering.

    This WSGI application expect to find the appropriate combination of the
    parameters tu,ts,tt,r,gu,gs,gt, or it generates a 400 error.
    """
    if env.get("REQUEST_METHOD") == "POST":
        s = env['wsgi.input'].read(int(env.get('CONTENT_LENGTH') or 0))
        fields = cgi.parse_qs(s)
    else:
        fields = cgi.parse_qs(env.get("QUERY_STRING",""))
    env["t4r.fields"] = fields
    env["t4r.base"] = base = \
      "%(wsgi.url_scheme)s://%(SERVER_NAME)s:%(SERVER_PORT)s/%(SCRIPT_NAME)s" \
      % env
    acc_lang = env["HTTP_ACCEPT_LANGUAGE"]

    if not check_fields(fields):
        msg = "\n".join(str(i) for i in fields.items())
        return error(env, start_response,
                     "400 Internal Server Error",
                     "Invalid or missing parameters. Got:\n" + msg)

    retrieve = (lambda key: fields.get(key, [""])[0] or None)
    tu = retrieve("tu"); ts = retrieve("ts"); tt = retrieve("tt")
    r = retrieve("r")
    gu = retrieve("gu"); gs = retrieve("gs"); gt = retrieve("gt")
    logstream = env.get("wsgi.errors")

    if tu:
        tu = smart_urljoin(base, tu)
    else:
        tu = make_typed_stream(ts, tt)
    if gu:
        gu = smart_urljoin(base, gu)
    elif gs:
        gu = make_typed_stream(gs, gt)

    global_t4r = {
        "fields": fields,
        "base": base,
        "accept-language": acc_lang,
    }
    if _allowed_requests is not None:
        global_t4r["allowed-requests"] = _allowed_requests
    globals = {"t4r": global_t4r}
    out = render_template(tu, r, gu, globals, logstream)
    start_response("200 OK", out.info().items())
    return out

def get_allowed_requests():
    return _allowed_requests

def set_allowed_requests(value):
    global _allowed_requests
    _allowed_requests = value

_allowed_requests = None

def check_fields(fields):
    return ("tu" in fields or ("ts" in fields and "tt" in fields)) \
       and "r" in fields \
       and ("gs" not in fields or "gt" in fields) \
       and ("gt" not in fields or "gs" in fields)

def error(env, start_response, code, message):
        if isinstance(message, unicode):
            message = message.encode("utf-8")
        assert isinstance(message, str)
        message += "\n\nFor support, contact pchampin at liris dot cnrs dot fr"
        start_response(code, [
            ("content-type", "text/plain;charset=utf-8"),
            ("content-length", str(len(message))),
        ])
        return [message]

def smart_urljoin(base, url):
    """
    This method provides security check and optimization.

    Security check: does not allow file: URLs to be accessed. Warning: some
    local files under the current directory can be opened (see below).

    Optimization: relative URLs without any '..' are first attempted as
    local files first, and as relative to base if they fail.
    """
    parsed_url = urlparse(url)
    if parsed_url[0] == "file":
        raise IOError("file: URLs are not allowed")
    if parsed_url == ('','',parsed_url[2],'','',''):
        # try local filename
        path = parsed_url[2].split("/")
        if ".." not in path:
            path = join(*path)
            if exists(path):
                path = pathname2url(abspath(path))
                return urlunparse(("file", None, path, None, None, None))
    # fall back to standard urljoin
    return urljoin(base, url)

#
