#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal
from gc import collect
from os.path import abspath, dirname, join
from sys import modules, path
from urllib import urlencode
from urllib2 import urlopen
from urlparse import urljoin, urlunparse

dir = dirname(__file__)
lib = join(dir, "..", "lib")
path.append(lib)
path.append(dir)

from t4r.context import QNameContext
from t4r.path import RdfPath, RdfPathException, RdfPathError
from t4r.graph import TalesGraph, parse_accept_language

from RDF import Model, Parser
from rdflib.Graph import Graph
from simpletal import simpleTALES

def test_parse_accept_language():
    for s,l1 in [
        ["fr,en-gb;q=.8,en", [("fr",), ("en", "gb"), ("en",),]],
        ["fr , en-gb ; q=.8 , en", [("fr",), ("en", "gb"), ("en",),]],
        ["fr , en-gb;q=.8 , en, *", [("fr",), ("en", "gb"), ("en",),("*",),]],
    ]:
        l2 = parse_accept_language(s)
        assert l1 == l2, (s, l1, l2)


class WithCheckMixin:
    """
    Define a useful check method used in most tests.

    This method assumed an instance variable tg, a TalesGraph.
    """
    def check(self, expected, tales,
                    ordered=False, startswith=False):
        r = self.tg._context.evaluate("g/%s" % tales)
        if isinstance(r, RdfPath):
            if isinstance(expected, bool):
                trial = not not r
            else:
                def C(seq, ordered):
                    lst = list(tuplize(i) for i in seq)
                    if ordered is False:
                        lst.sort()
                    return lst

                trial = C(( p._steps[0][0] for p in r ), ordered)
                expected = C(expected, ordered)
                #print "===", tales, [str(i) for i in r]
            assert trial == expected, "\n%r\n%r" % (trial, expected)
        elif startswith:
            assert r.startswith(expected), "\n%r\n%r" % (r, expected)
        else:
            assert r == expected, "\n%r\n%r" % (r, expected)


class BaseTestPathEvaluation(WithCheckMixin):
    """
    This base class is backend independant.
    It should be subclassed for each backend.
    """

    def test_test(self):
        assert self.tg._context.evaluate("g") is self.tg
        assert self.tg._context.evaluate("g/size") == self.tg.size()
        for p,u in namespaces:
            assert self.tg._context.evaluate("t4rns:%s" % p) == u

    def test_pending_or(self):
        err = RdfPathError("unfinished path: g/ex:jd/or")
        self.check(err, "ex:jd/or")

    def test_first_step(self):
        self.check([["+", ex.doc1],], "ex:doc1")
        self.check([["+", ex.not_in_the_graph],], "ex:not_in_the_graph")

    def test_2_steps(self):
        self.check([["+", foaf.Document], ["+", ex.WebPage],],
                   "ex:doc1/rdf:type", )
        self.check([["L", "John's homepage", "en", None],
                    ["L", "La page de John", "fr", None],
                   ], "ex:doc1/dc:title")

    def test_3_steps(self):
        self.check([["+", foaf.Person],], "ex:jd/foaf:knows/rdf:type")

    def test_inverse_step(self):
        self.check([["+", ex.doc1], ["+", ex.doc2],],
                   "foaf:Document/rdf:type:-")

    def test_with_initial_or(self):
        self.check([["+", ex.doc1], ["+", ex.doc2]], "ex:doc1/or/ex:doc2")
        self.check([["L", "John's homepage", "en", None],
                    ["L", "La page de John", "fr", None],
                    ["L", "Another document", None, xsd.string],
                   ], "ex:doc1/or/ex:doc2/dc:title")

    def test_with_intermediate_or(self):
        self.check([["+", ex.doc1],["+", ex.doc2]],
                   "ex:jd/foaf:homepage/or/foaf:made")
        self.check([["L", "John's homepage", "en", None],
                    ["L", "La page de John", "fr", None],
                    ["L", "Another document", None, xsd.string],
                   ], "ex:jd/foaf:homepage/or/foaf:made/dc:title")

    def test_with_mixed_or(self):
        self.check([["+", ex.doc1],["+", ex.doc2]],
                   "ex:jd/foaf:made/or/dc:creator:-")
        self.check([["L", "John's homepage", "en", None],
                    ["L", "La page de John", "fr", None],
                    ["L", "Another document", None, xsd.string],
                   ], "ex:jd/foaf:made/or/dc:creator:-/dc:title")

    def test_multiple_or(self):
        self.check([["+", foaf.Document],
                    ["+", foaf.Person],
                    ["+", ex.WebPage],
                   ],
                   "foaf:Document/or/foaf:Person/or/ex:WebPage")
        self.check([["+", ex.doc1],
                    ["+", ex.doc2],
                    ["L", "John Doe", None, None],
                   ],
                   "ex:jd/foaf:name/or/foaf:made/or/foaf:homepage")

    def test_several_or(self):
        self.check([["L", "John's homepage", "en", None],
                    ["L", "La page de John", "fr", None],
                    ["L", "Another document", None, xsd.string],
                    ["+", ex.doc2],
                   ],
                   "ex:jd/foaf:made/or/dc:creator:-/dc:title/or/rdfs:seeAlso")

    def test_indirection(self):
        self.tg._context.setLocal("COMPLEX_PATH", "foaf:made/or/dc:creator:-")
        self.check([["L", "John's homepage", "en", None],
                    ["L", "La page de John", "fr", None],
                    ["L", "Another document", None, xsd.string],
                    ["+", ex.doc2],
                   ],
                   "ex:jd/?COMPLEX_PATH/dc:title/or/rdfs:seeAlso")

    def test_no_duplicates(self):
        self.check([["+", ex.doc1],],
                   "ex:jd/dc:creator:-/or/foaf:homepage")

    def test_op_datatype(self):
        self.check([["+", xsd.string],], "ex:doc2/dc:title/datatype")
        self.check([], "ex:doc1/dc:title/datatype")
        self.check([], "ex:doc1/datatype")
        self.check([["+", xsd.string],],
                   "ex:doc2/dc:title/or/dc:creator/datatype")

    def test_op_iterSeq(self):
        self.check([["+", ex.item1], ["+", ex.item2], ["+", ex.item3],],
                   "ex:collections/ex:hasSeq/iterSeq", ordered=1)
        self.check([], "ex:jd/iterSeq")

    def test_op_iterList(self):
        self.check([["+", ex.item1], ["+", ex.item2], ["+", ex.item3],],
                   "ex:collections/ex:hasList/iterList", ordered=1)
        self.check([], "ex:jd/iterList")

    def test_op_any(self):
        self.check([["+", ex.item1],],
                   "ex:collections/ex:hasList/iterList/any", ordered=1)
        self.check([["+", ex.jd],], "ex:jd/any")
        self.check([["+", ex.doc1],], "ex:doc1/or/ex:doc2/any")
        try: self.check([], "ex:jd/dc:creator/any")
        except simpleTALES.PathNotFoundException: pass
        else: assert 0, "PathNotFoundException expected"

    def test_op_URIRefs_literals_blanks(self):
        path = "ex:jd/dc:creator:-/or/foaf:name/or/foaf:made/or/foaf:knows"
        self.check([["+", ex.doc1], ["+", ex.doc2]], path + "/URIRefs")
        self.check([["L", "John Doe", None, None]], path + "/literals")
        self.check(True, path + "/blanks/isBlank")

    def test_op_selectLang(self):
        path1 = "ex:txt1/iterSeq/selectLang"
        path2 = "ex:txt2/iterSeq/selectLang"
        all1 = [["L", "Text", None, None],
                ["L", "Text", "en", None],
                ["L", "Texte", "fr", None],
                ["L", "Texte", "fr-ca", None],
               ]
        all2 = [["L", "Texte", "fr-ca", None],
                ["L", "Texte", "fr-fr", None],
                ["L", "Le texte", "fr-fr", None],
                ["L", "Text", "en", None],
               ]

        self.check(all1, path1)
        self.check(all2, path2)

        env = self.tg._context.globals["t4r"] = {}
        def reset_env(env, accept_language):
            env[".parsed-accept-language"] = None
            env["accept-language"] = accept_language

        reset_env(env, "it")
        self.check([], path1) # it
        self.check([], path2) # it

        reset_env(env, "fr-fr")
        self.check([], path1) # fr-fr
        self.check([["L", "Texte", "fr-fr", None],
                    ["L", "Le texte", "fr-fr", None]], path2) # fr-fr

        reset_env(env, "it, fr-fr ; q=.8")
        self.check([], path1) # it,fr-fr
        self.check([["L", "Texte", "fr-fr", None],
                    ["L", "Le texte", "fr-fr", None]], path2) # it,fr-fr

        reset_env(env, "fr-be")
        self.check([], path1) # fr-be
        self.check([], path2) # fr-be

        reset_env(env, "fr")
        self.check([["L", "Texte", "fr", None]], path1) # fr
        self.check([["L", "Texte", "fr-ca", None],
                    ["L", "Texte", "fr-fr", None],
                    ["L", "Le texte", "fr-fr", None]], path2) # fr

        reset_env(env, "fr-fr, fr ; q=.8")
        self.check([["L", "Texte", "fr", None]], path1) # fr-fr,fr
        self.check([["L", "Texte", "fr-fr", None],
                    ["L", "Le texte", "fr-fr", None]], path2) # fr-fr,fr

        reset_env(env, "fr-fr, en ; q=.5, *")
        self.check([["L", "Text", "en", None]], path1) # fr-fr,en,*
        self.check([["L", "Texte", "fr-fr", None],
                    ["L", "Le texte", "fr-fr", None]], path2) # fr-fr,en,*

        reset_env(env, "fr-fr,fr,en,*")
        self.check([["L", "Texte", "fr", None]], path1) # fr-fr,fr,en,*
        self.check([["L", "Texte", "fr-fr", None],
                    ["L", "Le texte", "fr-fr", None]], path2) # fr-fr,fr,en,*

        reset_env(env, "it,*")
        self.check(all1, path1) # it,*
        self.check(all2, path2) # it,*

    def test_op_convert(self):
        self.check(1, "ex:data/ex:anInt/convert")
        self.check(2, "ex:data/ex:anInteger/convert")
        self.check(3.0, "ex:data/ex:aFloat/convert")
        self.check(4.0, "ex:data/ex:aDouble/convert")
        self.check(5, "ex:data/ex:aDecimal1/convert")
        self.check(Decimal("6.0"), "ex:data/ex:aDecimal2/convert")
        self.check(u"\xe9", "ex:data/ex:aString/convert")
        self.check(u"8", "ex:data/ex:anUntypedLiteral/convert")
        # try with multiple nodes
        try: self.check(None, "ex:doc1/dc:title/convert")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/convert")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with URIRef node
        try: self.check(None, "ex:doc1/convert")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/convert")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with unsupported datatype
        try: self.check(None, "ex:data/ex:aDate/convert")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_id(self):
        global data_ttl_url
        self.check(u"jd", "ex:jd/id") # slash
        self.check(u"type", "rdf:type/id") # hash
        self.check(u"def", "ex:doc2/ex:opaqueUri/id") # colon
        # can not easily test bnode, since we do not know their ID:
        try: self.check(u"", "ex:jd/foaf:knows/id", startswith=1)
        except RdfPathException:
            assert 0, "Unexpected exception, should work with blank node"
        # try with multiple nodes
        try: self.check(0,"ex:jd/foaf:homepage/or/foaf:knows/id")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/id")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with Literal node
        try: self.check(None, "ex:jd/foaf:name/id")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_isURIRef(self):
        self.check(True, "ex:jd/isURIRef")
        self.check(False, "ex:jd/foaf:name/isURIRef")
        self.check(False, "ex:jd/foaf:knows/isURIRef")
        # try with multiple nodes
        try: self.check(0,"ex:jd/foaf:name/or/foaf:made/or/foaf:knows/isURIRef")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/isURIRef")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_isLiteral(self):
        self.check(False, "ex:jd/isLiteral")
        self.check(True, "ex:jd/foaf:name/isLiteral")
        self.check(False, "ex:jd/foaf:knows/isLiteral")
        # try with multiple nodes
        try:self.check(0,"ex:jd/foaf:name/or/foaf:made/or/foaf:knows/isLiteral")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/isLiteral")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_isBlank(self):
        self.check(False, "ex:jd/isBlank")
        self.check(False, "ex:jd/foaf:name/isBlank")
        self.check(True, "ex:jd/foaf:knows/isBlank")
        # try with multiple nodes
        try: self.check(0,"ex:jd/foaf:name/or/foaf:made/or/foaf:knows/isBlank")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/isBlank")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_n3(self):
        global data_ttl_url
        self.check(u"ex:jd", "ex:jd/n3")
        self.check(u"<%s>" % data_ttl_url, "ex:jd/rdfs:seeAlso/n3")
        self.check(u'"John Doe"', "ex:jd/foaf:name/n3")
        self.check(u'"1"^^xsd:int', "ex:data/ex:anInt/n3")
        self.check(u'"8"@fr', "ex:data/ex:anUntypedLiteral/n3")
        self.check(u"_:", "ex:jd/foaf:knows/n3", startswith=1)
        # can not easily test bnode, since we do not know their ID
        self.check(u'"John Doe"', "ex:jd/foaf:name/n3")
        # try with multiple nodes
        try: self.check(0,"ex:jd/foaf:name/or/foaf:made/or/foaf:knows/n3")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/n3")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_ns(self):
        global data_ttl_url
        self.check([["+", ex._uri]], "ex:jd/ns") # slash
        self.check([["+", rdf._uri]], "rdf:type/ns") # hash
        self.check([["+", "abc:"]], "ex:doc2/ex:opaqueUri/ns") # colon
        # try with multiple nodes
        try: self.check(0,"ex:jd/foaf:homepage/or/foaf:knows/ns")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/ns")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with Literal node
        try: self.check(None, "ex:jd/foaf:name/ns")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/ns")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_uri(self):
        self.check(u"http://example.org/terms/jd", "ex:jd/uri")
        self.check(u"http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
                   "rdf:type/uri")
        # try with multiple nodes
        try: self.check(0,"ex:doc1/or/ex:doc2/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with Literal node
        try: self.check(None, "ex:jd/foaf:name/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_text(self):
        self.check(u"http://example.org/terms/jd", "ex:jd/uri")
        self.check(u"http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
                   "rdf:type/uri")
        # try with multiple nodes
        try: self.check(0,"ex:doc1/or/ex:doc2/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with Literal node
        try: self.check(None, "ex:jd/foaf:name/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_lang(self):
        self.check(u"John Doe", "ex:jd/foaf:name/text")
        # try with multiple nodes
        try: self.check(None, "ex:doc1/dc:title/text")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/text")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with URIRef node
        try: self.check(None, "ex:doc1/text")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/text")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_asPath(self):
        self.check("ex:jd",
                   "ex:jd/asPath")
        self.check("ex:doc1/or/ex:doc2",
                   "ex:doc1/or/ex:doc2/asPath")
        self.check("ex:item1/or/ex:item2/or/ex:item3",
                   "ex:collections/ex:hasSeq/iterSeq/asPath")
        self.check("ex:doc2",
                   "ex:jd/foaf:name/or/foaf:made/or/foaf:knows/asPath")

    def test_op_asInveresePath(self):
        self.check("ex:jd:-",
                   "ex:jd/asInversePath")
        self.check("ex:doc1:-/or/ex:doc2:-",
                   "ex:doc1/or/ex:doc2/asInversePath")
        self.check("ex:item1:-/or/ex:item2:-/or/ex:item3:-",
                   "ex:collections/ex:hasSeq/iterSeq/asInversePath")
        self.check("ex:doc2:-",
                   "ex:jd/foaf:name/or/foaf:made/or/foaf:knows/asInversePath")

    def test_op_count(self):
        self.check(3, "ex:collections/ex:hasSeq/iterSeq/count")
        self.check(2, "ex:jd/foaf:made/or/dc:creator:-/count")
        self.check(1, "ex:jd/foaf:made/count")
        self.check(0, "ex:jd/foaf:made:-/count")

    def test_op_link_r(self):
        env = self.tg._context.globals["t4r"] = {}
        fields = env["fields"] = {}
        fields["tu"] = ["./the_template",]
        fields["tt"] = ["text/html;charset=utf-8"]
        # we are testing in the case of a resource template
        fields["r"] = ["http://champin.net/foaf.rdf#pa"]
        # try without a graph
        x = lambda r: "?"+urlencode([("ts", ""), ("tu", fields["tu"][0]),
                                   ("tt", fields["tt"][0]), ("r", r)])
        self.check(x(ex.jd), "ex:jd/link")
        self.check(x(rdf.type), "rdf:type/link")
        # try with a graph URI
        fields["gu"] = ["./the_graph",]
        fields["gt"] = ["n3",]
        x = lambda r: "?"+urlencode([("ts", ""), ("tu", fields["tu"][0]),
                                   ("tt", fields["tt"][0]),
                                   ("gs", ""), ("gu", fields["gu"][0]),
                                   ("gt", "n3"), ("r", r)])
        self.check(x(ex.jd), "ex:jd/link")
        self.check(x(rdf.type), "rdf:type/link")
        # try with a graph URI
        del fields["gu"]
        fields["gs"] = ["# some n3",]
        fields["gt"] = ["n3",]
        x = lambda r: "?"+urlencode([("ts", ""), ("tu", fields["tu"][0]),
                                   ("tt", fields["tt"][0]),
                                   ("gs", fields["gs"][0]), ("gu", ""),
                                   ("gt", "n3"), ("r", r)])
        self.check(x(ex.jd), "ex:jd/link")
        self.check(x(rdf.type), "rdf:type/link")
        # try with multiple nodes
        try: self.check(0,"ex:doc1/or/ex:doc2/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with Literal node
        try: self.check(None, "ex:jd/foaf:name/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_link_g(self):
        env = self.tg._context.globals["t4r"] = {}
        fields = env["fields"] = {}
        fields["tu"] = ["./the_template",]
        fields["tt"] = ["text/html;charset=utf-8"]
        fields["gu"] = ["./the_graph",]
        fields["gt"] = ["n3",]
        # try without a resource
        x = lambda r: "?"+urlencode([("ts", ""), ("tu", fields["tu"][0]),
                                   ("tt", fields["tt"][0]),
                                   ("gs", ""), ("gu", r),
                                   ("gt", "")])
        self.check(x(ex.jd), "ex:jd/link")
        self.check(x(rdf.type), "rdf:type/link")
        # try with multiple nodes
        try: self.check(0,"ex:doc1/or/ex:doc2/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with no node
        try: self.check(None, "ex:jd/dc:creator/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with Literal node
        try: self.check(None, "ex:jd/foaf:name/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"
        # try with blank node
        try: self.check(None, "ex:jd/foaf:knows/uri")
        except RdfPathException: pass
        else: assert 0, "Exception expected"

    def test_op_contains(self):
        self.check(True,
            "ex:jd/foaf:made/or/dc:creator:-/contains/ex:doc1/or/ex:doc2")
        self.check(False,
            "ex:jd/foaf:knows/contains/ex:doc1/dc:creator:-")

    def test_op_renderWith(self):
        dir = urljoin(data_ttl_url, ".")
        r = self.tg._context.addGlobal("t4rns:dir", dir.decode())
        self.check("name: John Doe",
            "ex:jd/renderWith/dir:module.xml")
        self.check([["+", foaf.name]],
            "foaf:name")

    def test_longer_path(self):
        self.check(u"john doe", "ex:jd/foaf:name/convert/lower")

    def test_truth_value(self):
        self.check(True, "ex:collections/ex:hasSeq/iterSeq")
        self.check(True, "ex:jd/foaf:made/or/dc:creator:-")
        self.check(True, "ex:jd/foaf:made")
        self.check(False, "ex:jd/foaf:made:-")

    def test_unrecognized_step(self):
        err = RdfPathError("invalid path step: g/ex:jd/foo")
        self.check(err, "ex:jd/foo")

    def test_unrecognized_steps(self):
        err = RdfPathError("invalid path step: g/ex:jd/foo")
        self.check(err, "ex:jd/foo/bar")

    def test_g(self):
        r = self.tg._context.evaluate("g")
        assert r is self.tg

    def test_iter_g(self):
        lst = list(self.tg)
        assert len(lst) == self.tg.size()
        for t in lst:
            for k in ["s", "p", "o", "subject", "predicate", "object"]:
                assert k in t
                assert isinstance(t[k], RdfPath)
                assert len(list(t[k])) == 1

class BaseTestFactories:
    """
    This base class is backend independant.
    It should be subclassed for each backend.
    """
    def test_from_url(self):
        global data_ttl_url
        c = QNameContext()
        g = TalesGraph.from_url(data_ttl_url, c)
        assert self.backend in g.__class__.__name__
        assert g.size() > 0

    def test_from_stream(self):
        global data_ttl_url
        c = QNameContext()
        f = urlopen(data_ttl_url)
        g = TalesGraph.from_stream(f, c)
        assert self.backend in g.__class__.__name__
        assert g.size() > 0

class BaseTestFyn(WithCheckMixin):
    """
    This base class is backend independant.
    It should be subclassed for each backend.
    """
    def test_fyn(self):
        global data2_ttl_url
        c = QNameContext()
        add_t4r_namespaces(c)
        c.addGlobal("t4rns:data2", "%s#" % data2_ttl_url)
        tg = self.tg = TalesGraph.from_follow_your_nose(c)
        c.addGlobal("g", tg)
        self.check([["+", ex.jd],], "data2:paul/foaf:knows", )
        self.check([["L", u"Jane Smyth", None, None],],
                   "data2:paul/foaf:knows/foaf:knows/foaf:name", )


try:
    import RDF
    make_flakes_happy = RDF

    class TestRedlandPathEvaluation(BaseTestPathEvaluation):
        def setUp(self):
            global data_ttl_url
            m = Model()
            p = Parser(mime_type="application/x-turtle")
            p.parse_into_model(m, data_ttl_url)
            c = QNameContext()
            add_t4r_namespaces(c)
            tg = TalesGraph.from_redland(m, c)
            c.addGlobal("g", tg)
            self.tg = tg

    class DisableRdflibMixin(object):
        def setUp(self):
            self.backend = "Redland"
            try:
                # inhibit redland if present
                import rdflib
                self.rdflib = rdflib
                modules["rdflib"] = None
            except ImportError:
                self.rdflib = None

        def tearDown(self):
            if self.rdflib:
                modules["rdflib"] = self.rdflib

    class TestRedlandFactories(BaseTestFactories, DisableRdflibMixin):
        pass

    class TestRedlandFyn(BaseTestFyn, DisableRdflibMixin):
        pass

    #

except ImportError:
    pass


try:
    import rdflib
    make_flakes_happy = rdflib

    class TestRdfLibPathEvaluation(BaseTestPathEvaluation):
        def setUp(self):
            global data_ttl_url
            g = Graph()
            g.load(data_ttl_url, format="n3")
            c = QNameContext()
            add_t4r_namespaces(c)
            tg = TalesGraph.from_rdflib(g, c)
            c.addGlobal("g", tg)
            self.tg = tg

    class DisableRedlandMixin(object):
        def setUp(self):
            self.backend = "Rdflib"
            try:
                # inhibit redland if present
                import RDF
                self.RDF = RDF
                collect()
                modules["RDF"] = None
            except ImportError:
                self.RDF = None

        def tearDown(self):
            if self.RDF:
                modules["RDF"] = self.RDF

    class TestRdflibFactories(BaseTestFactories, DisableRedlandMixin):
        pass

    class TestRdflibFyn(BaseTestFyn, DisableRedlandMixin):
        pass

    #

except ImportError:
    pass



# ================ Test machinery ================

from uris_4_tests import namespaces, add_t4r_namespaces, add_global_generators

import log_4_test
log_4_test # just to make pyflakes happy

ex, foaf, xsd, rdf = (None,)*4 # just to make pyflakes happy
add_global_generators(globals()) # this adds NS variables in the global scope

data_ttl_url = urlunparse(
    ("file", None, abspath(join(dir, "data.ttl")), None, None, None))

data2_ttl_url = urlunparse(
    ("file", None, abspath(join(dir, "data2.ttl")), None, None, None))



def tuplize(arg):
    """
    Recursivly change lists into tuples.
    """
    if isinstance(arg, list):
        return tuple( tuplize(i) for i in arg )
    else:
        return arg

#
