#    This file is part of Tal4Rdf <http://champin.net/t4r/>
#    Copyright (C) 2008 Pierre-Antoine Champin <pchampin@liris.cnrs.fr>
#
#    Tal4Rdf is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Tal4Rdf is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with Tal4Rdf.  If not, see <http://www.gnu.org/licenses/>.

from os.path import dirname, join
import sys

dir = dirname(__file__)
lib = join(dir, "..", "lib")
sys.path.append(lib)
sys.path.append(dir)

from t4r.path import RdfPath, RdfPathError, \
                     DATA_OPERATORS, NODE_OPERATORS, BINARY_OPERATORS

class TestCaseParsing:
    def setUp(self):
        #print >> sys.stderr, "=== setting up"
        self.path = RdfPath.for_graph(DummyGraph(), "ex:john")

    def tearDown(self):
        #print >> sys.stderr, "=== tearing down"
        pass

    def test_first_step(self):
        trial = self.path._steps
        expected = [[["+", ex.john],],]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_2_steps(self):
        trial = self.path["foaf:knows"]._steps
        expected = [
          [["+", ex.john],],
          [["+", foaf.knows],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_3_steps(self):
        trial = self.path["foaf:knows"]["ex:paul"]._steps
        expected = [
          [["+", ex.john],],
          [["+", foaf.knows],],
          [["+", ex.paul],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_inverse_step(self):
        trial = self.path["foaf:knows:-"]["ex:paul"]._steps
        expected = [
          [["+", ex.john],],
          [["-", foaf.knows],],
          [["+", ex.paul],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_with_initial_or(self):
        trial = self.path["or"]["ex:paul"]["foaf:knows"]._steps
        expected = [
          [["+", ex.john], ["+", ex.paul],],
          [["+", foaf.knows],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_with_intermediate_or(self):
        trial = self.path["foaf:knows"]["or"]["ex:playswith"]["ex:paul"]._steps
        expected = [
          [["+", ex.john],],
          [["+", foaf.knows], ["+", ex.playswith],],
          [["+", ex.paul],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_with_mixed_or(self):
        trial = self.path["foaf:knows"]["or"]["ex:playswith:-"]["ex:paul"]._steps
        expected = [
          [["+", ex.john],],
          [["+", foaf.knows], ["-", ex.playswith],],
          [["+", ex.paul],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_with_multiple_or(self):
        trial = self.path["foaf:knows"]["or"]["ex:playswith"]["or"]["ex:dislikes"]["ex:paul"]._steps
        expected = [
          [["+", ex.john],],
          [["+", foaf.knows], ["+", ex.playswith], ["+", ex.dislikes],],
          [["+", ex.paul],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    def test_with_several_or(self):
        trial = self.path["foaf:knows"]["or"]["ex:playswith"]["ex:paul"]["or"]["ex:george"]._steps
        expected = [
          [["+", ex.john],],
          [["+", foaf.knows], ["+", ex.playswith]],
          [["+", ex.paul], ["+", ex.george],],
        ]
        assert trial == expected, "\n%r\n%r" % (trial, expected)

    for data_op in DATA_OPERATORS:
        # generate one test method for each data operator
        def _test_data_op_X(self, data_op=data_op):
            p = self.path["foaf:knows"]["or"]["ex:playswith"][data_op]
            trial = p._steps
            expected = [
              [["+", ex.john],],
              [["+", foaf.knows], ["+", ex.playswith]],
            ]
            assert trial == expected, "\n%r\n%r" % (trial, expected)
            assert self.path._operator == data_op
            assert self.path._type == DATA_OPERATORS[data_op]
            try: p["ex:paul"]
            except AssertionError: pass
            else: assert "Expected error after .../%s/ex:paul" % data_op

        # some magic to add the method with the right name
        name = "test_data_op_%s" % data_op
        _test_data_op_X.__name__ = name
        locals()[name] = _test_data_op_X
        del _test_data_op_X

    for node_op in NODE_OPERATORS:
        # generate one test method for each node operator
        def _test_node_op_X(self, node_op=node_op):
            trial = self.path["foaf:knows"]["or"]["ex:playswith"][node_op]["ex:paul"]._steps
            expected = [
              [["+", ex.john],],
              [["+", foaf.knows], ["+", ex.playswith]],
              [["K", node_op],],
              [["+", ex.paul],],
            ]
            assert trial == expected, "\n%r\n%r" % (trial, expected)

        # some magic to add the method with the right name
        name = "test_node_op_%s" % node_op
        _test_node_op_X.__name__ = name
        locals()[name] = _test_node_op_X
        del _test_node_op_X

    for bin_op in BINARY_OPERATORS:
        # generate one test method for each bin operator
        def _test_bin_op_X(self, bin_op=bin_op):
            p = self.path["foaf:knows"]["or"]["ex:playswith"][bin_op]["ex:paul"]["foaf:knows"]
            trial = p._steps
            expected = [
              [["+", ex.paul],],
              [["+", foaf.knows],],
            ]
            assert trial == expected, "\n%r\n%r" % (trial, expected)
            trial1 = p._steps1
            expected1 = [
              [["+", ex.john],],
              [["+", foaf.knows], ["+", ex.playswith]],
            ]
            assert trial1 == expected1, "\n%r\n%r" % (trial1, expected1)
            assert p._operator == bin_op
            assert p._type == BINARY_OPERATORS[bin_op]
        # some magic to add the method with the right name
        name = "test_bin_op_%s" % bin_op
        _test_bin_op_X.__name__ = name
        locals()[name] = _test_bin_op_X
        del _test_bin_op_X

    def test_undeclared_namespace(self):
        expected = RdfPathError("undeclared namespace: foo")
        try:
            self.path["foo:bar"]
        except RdfPathError, trial:
            assert trial == expected, "\n%r\n%r" % (trial, expected)
        else:
            assert 0, "RdfPathError expected"

    def test_unknown_suffix(self):
        expected = RdfPathError("unknown xqname suffix: foo")
        try:
            self.path["foaf:knows:foo"]
        except RdfPathError, trial:
            assert trial == expected, "\n%r\n%r" % (trial, expected)
        else:
            assert 0, "RdfPathError expected"

def test_mutable():
    # for this test, DummyGraph is too dummy
    # so we have dependancies on .graph and .context
    from t4r.graph import TalesGraph
    from t4r.context import QNameContext
    c = QNameContext()
    c.addGlobal("t4rns:foaf", getattr(foaf,""))
    c.addGlobal("t4rns:ex", getattr(ex,""))
    f = open(join(dir, "data.ttl"))
    g = TalesGraph.from_stream(f, c)
    f.close()
    p = RdfPath.for_graph(g, "ex:jd")

    p.value((0,[]))
    assert not p._mutable
    assert len(p._steps) == 1, p._steps # BEFORE Q
    q = p["foaf:name"]
    assert q is not p
    assert not p._mutable
    assert len(p._steps) == 1, p._steps # AFTER Q
    assert q._mutable
    q.value((0,[]))
    assert not q._mutable
    assert len(q._steps) == 1, q._steps # Q EVALUATED

def test_strict_display():
    # for this test, DummyGraph is too dummy
    # so we have dependancies on .graph and .context
    from t4r.graph import TalesGraph
    from t4r.context import QNameContext
    c = QNameContext()
    c.addGlobal("t4rns:foaf", getattr(foaf,""))
    c.addGlobal("t4rns:dc", getattr(dc,""))
    c.addGlobal("t4rns:ex", getattr(ex,""))
    f = open(join(dir, "data.ttl"))
    g = TalesGraph.from_stream(f, c)
    c.set_graph_and_resource(g, RdfPath.for_resource(g, "http://example.com"))
    f.close()

    p0 = c.evaluate("/ex:jd/dc:creator")
    p2 = c.evaluate("/ex:jd/foaf:depiction")
    uris = [ex.img1, ex.img2,]
    c.addGlobal("t4r:display", "strict")
    assert str(p0) == "{0 nodes}", repr(str(p0))
    assert str(p2) == "{2 nodes}", repr(str(p2))
    c.setLocal("t4r:display", "")
    assert str(p0) == "", repr(str(p0))
    assert str(p2) in uris, repr(str(p2))

def test_undeclared_namespace_in_first_qname():
    expected = RdfPathError("undeclared namespace: foo")
    try:
        RdfPath.for_graph(DummyGraph(), "foo:bar")
    except RdfPathError, trial:
        assert trial == expected, "\n%r\n%r" % (trial, expected)
    else:
        assert 0, "RdfPathError expected"

# ======== Test machinery ========

from uris_4_tests import add_global_generators

dc, ex, foaf, xsd = (None,)*4 # just to make pyflakes happy
add_global_generators(globals()) # this adds NS variables in the global scope

class DummyGraph(object):
    _context = None # not used for the moment
    def _qname2uriref(self, prefix, suffix):
        gen = globals().get(prefix)
        if gen is None:
            return None
        else:
            return getattr(gen, suffix)

import log_4_test
log_4_test # just to make pyflakes happy

#
